unit Tokeniser;
(* Simple tokeniser and two handy splitting functions

  ###############################
  ### Peter Fox               ###    Put in the public domain
  ### www.eminent.demon.co.uk ###    by the author.
  ### September 2004          ###
  ###############################

  (Added escape character 7 Sept 04)



  Example
  ============================================================
  | T := ttokeniser.create('=','Rover=Dog|black|collie|7|||');
  | if T.anymore then memo.lines.add('['+t.nextbit+']');   // Get Rover
  | T.setsplitter('|');                                    // Change splitter
  | while t.anymore do begin                               // work through rest
  |   memo.lines.add('['+t.nextbit+']');                   // bits about Rover
  | end;
  |
  | T.initialise('~-.0123456789','12.34 up to 56.78');     // reuse tokeniser
  | while t.anymore do begin                               // result is just
  |   memo.lines.add('['+t.nextbit+']');                   // the two numbers
  | end;
  |
  | T.free;                                                // don't forget!


  Splitter flexibility
  ============================================================
  ''      ... whitespace  (ascii <= 32)
  'a'     ... split if character a found
  'abc'   ... split if characters a or b or c found

  '-...'  ... prefix with minus sign to invert
  eg '-0123456789' will split on next non digit

  '~...'  ... prefix with tilde to squash splits
  eg  '~' treats '   ' as a single bit of whitespace character
  eg '~-0123456789' will break on any non digit and series of non digits
  (Compare splitting '12.34    56.78' with and without ~)
   ~ must preceede - if present


  Notes
  ============================================================
  There is notionally 'more' when the last char is a splitter
  Splitting ('1,2,3')  with ',' gives 1 2 3
  Splitting ('1,2,3,') with ',' gives 1 2 3 blank

  Splitting only works on single characters so if you want to
  extract say the minimum from 'MAX=198  MIN=124  AVG=155' you have
  to split on '~= ', find the bit 'MIN' then get the next bit as your number.

  An escape character (defaults to \, but can be set to '') is
  provided so that, for example if 'Peter\ Fox' wouldn't split on a space.


  Handy utility functions
  ============================================================
  Split   String --> array
  Join    Array --> String

*)

interface

uses
    Sysutils;

type
    TTokeniserMode = (tmWhite,tmSingle,tmMultiple);  // only used inside

    TTokeniser = class(Tobject)
      private
        fString    : string;
        fSplitter  : string;
        fIndex     : integer;
        fSplitMode : TTokeniserMode;
        fNegative  : boolean;         // true if we split an anything else
        fSquash    : boolean;         // true if we merge consecutive splitters found into 1
        function     isCharASplitter(var ix:integer):boolean;
      public
        Escape    : char;            // escape character Defaults to \
        constructor Create;                                   overload;
        constructor Create(const Splitter,aString:string);     overload;
        procedure   Initialise(const Splitter,aString:string);
        procedure   reset;
        function    AnyMore:boolean;
        function    NextBit:string;                             overload;
        function    NextBit(const Splitter:string):string;      overload;
        function    Remainder:string;
        procedure   SetSplitter(const Splitter:string);
        function    LastSplitChar:string;
        function    Position:integer;
    end;

// stand-alone functions
function Join(const ItemCount:integer;const Splitter:string; const StrArray:array of string):string;
function Split(const aString,Splitter:string; var StrArray:array of string):integer;
function divide(const Source:String; const targets:array of string;var parts:array of string):string;

implementation





function Join(const ItemCount:integer;const Splitter:string; const StrArray:array of string):string;
{ stick the first ITEMCOUNT number of elements from the array
  to make a string 'split' by the 'splitter'.
  eg   join(2,'!',['foo','bar','fox']) --> 'foo!bar'

  Converse to split()
}
var
  i : integer;
  r : string;
begin
  r := '';
  for i := 0 to Itemcount-1 do r := r + splitter + StrArray[i];
  if r <> '' then r := copy(r,2,length(r));
  result := r;
end;




function  Split(const aString,Splitter:string; var StrArray:array of string):integer;
{ split the given string into an array
  SPLITTER can be a simple character or more sophisticated
  See TTokeniser.SetSplitter for details
  The return value is the number of elements put into StrArray
}
var
  T : TTokeniser;
  i : integer;
begin
  i := 0;
  T := tTokeniser.create(Splitter,aString);
  TRY
    while (T.anymore and (i<= high(StrArray))) do begin
      StrArray[i] := T.nextbit;
      inc(i);
    end;
  FINALLY
    T.free;
  END;
  result := i;
end;






function  TTokeniser.AnyMore:boolean;
{ return true if there is more string left.
  NB  If the last splitting character was the last
      character of the source string then there is
      'more' even though it is a blank.
      Splitting ('1,2,3')  with ',' gives 1 2 3
      Splitting ('1,,3,')  with ',' gives 1 blank 3 blank
 }
begin
  result := (findex >= 0);
end;




constructor TTokeniser.Create(const Splitter,aString:string);
{ create ready to go }
begin
  create;
  initialise(splitter,aString);
end;




constructor TTokeniser.Create;
{ default constructor }
begin
  inherited create;
  fsplitter := '';
  fString := '';
  fIndex := -1;
  fNegative := false;
  fSquash := false;
  Escape := '\';
end;




procedure TTokeniser.Initialise(const Splitter,aString:string);
{ set up string and splitter but don't do anything yet }
begin
  reset;

  fString := aString;               // line up string
  if aString='' then fIndex := -1;  // no more!

  SetSplitter(Splitter);
end;




function  TTokeniser.isCharASplitter(var ix:integer):boolean;
{ return true if the character at given index is a splitter }
var
  c : char;
  hit : boolean;
begin
  hit := false;  // shut up compiler

  // make sure that this isn't following escape character
  if ix>1 then begin
    if fstring[ix-1]=Escape then begin
      result := false;
      exit;
    end;
  end;

  // test the character
  c := fString[ix];
  case fSplitMode of
  tmWhite     : hit := ord(c)<=32;          // white space & control characters
  tmSingle    : hit := (c = fsplitter);     // exact match
  tmMultiple  : hit := pos(c,fSplitter)>0;  // any
  end;{case}

  if fNegative then hit := not hit;      // possibly flip the result
  result := hit;
end;




function  TTokeniser.LastSplitChar:string;
{ return the character actually used to split.
  example
  |  T := TTokeniser.create('-0123456789','12.34/56.789');
  |  N := T.NextBit;
  |  if T.LastSplitChar = '.' then N := N + '.'+T.NextBit;
  |  (do something with N as a floating point number)
}
begin
  result := '';
  if findex < 1 then exit;                // nothing yet or off end
  if findex > length(fString) then exit;  // can never happen
  result := fString[findex];
end;




function  TTokeniser.NextBit(const Splitter:string):string;
{ split but with a different splitter }
begin
  SetSplitter(Splitter);
  result := nextbit(fSplitter);
end;




function  TTokeniser.NextBit:string;
{ 'same again' split }
var
  s,f,ls : integer;
begin
  result := '';
  s := 1 ;                    //shut up compiler
  if not AnyMore then exit;   // findex is -1 !

  // special case
  // very last character of string is a splitter
  // return a blank THEN say no more
  if fIndex = length(fString) then begin
    fIndex := -1;
    exit;
  end;

  // fIndex points to last place we found (or 0)
  // (BTW Strings are 1-indexed)
  // scan along the string until we get a splitter
  while findex >= 0 do begin
    s := fIndex+1;
    f := s;
    ls := length(fString);
    while f <= ls do begin
      if isCharASplitter(f) then begin  // found a split
        break;
      end;
      inc(f);
    end;
    if f>ls then fIndex := -1 else fIndex := f;
    if f-s > 0 then break;
    if not fSquash then break;
  end;

  // cut out the bit of string we want and update index
  // special case when f>ls (f=ls ... still 'more')
  result := copy(fString,s,f-s);
end;




function    TTokeniser.Position:integer;
{ report where the last split was found
  -1 = not initialised or past end
  0  = not started
  1  = 1st character
  eg  after splitting 'foo+bar' with '+' this would return 4
}
begin
  result := fIndex;
end;




function  TTokeniser.Remainder:string;
{ get the rest of the string
  Note : does not affect the index so remainder followed by nextbit will work fine
}
begin
  result := '';
  if not AnyMore then exit;   // findex is -1 !
  result := copy(fString,fIndex+1,length(fString));
end;




procedure TTokeniser.reset;
{ set index back to start }
begin
  fIndex := 0;                      // start position
end;




procedure TTokeniser.SetSplitter(const Splitter:string);
{ set up splitter CHARACTER(S)
  ''     ... white space
  '-'    ... not white space (!)
  'a'    ... character 'a'
  'abc'  ... any of a,b or c
  '-xyz' ... anything but x,y or z
  Prefix with tilde (~) to treat consecutive splitters as one
  eg  '~o'on 'book' --> b k whereas just 'o' --> b blank k

}
var
  c,s : string;
begin
  s := splitter;
  fSplitter  := s;
  fSplitMode := tmWhite;
  fNegative  := false;
  fSquash    := false;

  // now work out special conditions
  if s = '' then exit;     // white space

  c := s[1];              // squash ?
  fSquash := (c = '~');
  if fsquash then begin
    s := copy(s,2,length(s));
    fSplitter := s;
  end;
  if s = '' then exit;

  c := s[1];              // negative ?
  fNegative := (c = '-');
  if fnegative then begin
    s := copy(s,2,length(s));
    fSplitter := s;
  end;
  if s = '' then exit;

  // single character or multiple
  if length(fSplitter)=1 then fsplitmode := tmSingle else fsplitmode := tmMultiple;
end;

function divide(const Source:String; const targets:array of string;var parts:array of string):string;
{ this divides the source up by STRING tokens
  whitespace is assumed to be the lower level splitter
  eg
  | srce := 'SELECT something FROM else ORDER BY sort';
  | targetsarray := ['SELECT','FROM','WITH','ORDER','BY'];
  | beforeselect := divide(srce,targetsarray,partsarray);
  | fromclause   := partsarray[1];

  The function return value is anything before a target is found
  The results in parts match are the bits that follow the targets
  Targets are case insensitive
}
var
  T : TTokeniser;
  Bits,WorkingTargets : array of string;
  nBits,i,tokix,partix,nTargets,targetix : integer;
  beginning,s : string;
begin
  beginning := '';

  // split source into atomic tokens
  Setlength(Bits,(length(Source)+2) div 2);
  nBits := Split(trim(Source),'',Bits);

  // we want to crash case on our targets
  nTargets := length(targets);
  SetLength(WorkingTargets,nTargets);
  for i := 0 to nTargets-1 do WorkingTargets[i] := lowercase(targets[i]);
  // make sure the results array is clear
  for i := 0 to length(parts)-1 do parts[i] := '';

  // step through looking for the targets we want
  partix := -1;
  for tokix := 0 to nBits -1 do begin
    s := lowercase(bits[tokix]);
    // is this a target?
    targetix := -1;
    for i := 0 to nTargets-1 do begin
      if s = workingTargets[i] then begin
        targetix := i;
        break;
      end;
    end;
    // if so then move the place to put results
    if targetix>-1 then begin
      partix := targetix;
    end else begin      // if not then add to results
      if partix < 0 then beginning := beginning + ' ' + bits[tokix]
                    else parts[partix] := parts[partix] + ' ' + bits[tokix];
    end;
  end;
  result := beginning;
end;



end.
